
import std.stdio;
import std.format;
import std.string;
import core.stdc.errno;
import core.stdc.string;
import core.sys.posix.unistd;
import core.sys.freebsd.config;
import core.sys.freebsd.sys.event;
import msgq;
import esrc;

final class EventQueue {

    static private EventQueue _eq;

    private int id;
    private bool done;
    private MessageQueue mq;

    private this() {
        id = kqueue();
        assert(id > 0, format("kqueue() failed : %s", fromStringz(strerror(errno))));
        mq = MessageQueue.getInstance();
        writefln("### FreeBSD_version = %s", __FreeBSD_version);
        writefln("### sizeof(kevent_t) = %s", kevent_t.sizeof);
    }

    private ~this() {
        close(id);
    }

    static EventQueue getInstance() {
        if (_eq is null)
            _eq = new EventQueue();
        return _eq;
    }

/*
struct kevent_t
{
    uintptr_t    ident; // identifier for this event
    short       filter; // filter for event
    ushort       flags;
    uint        fflags;
    intptr_t      data;
    void        *udata; // opaque user data identifier
}
*/
    void enableEvent(short evtype, EventSource es) {
        kevent_t ke;
        //EV_SET(&ke, es.id, evtype, EV_ADD | EV_ENABLE | EV_DISPATCH, 0, es.evdata, cast(void*)es, [0, 0, 0, 0]);
        EV_SET(&ke, es.id, evtype, EV_ADD | EV_ENABLE | EV_DISPATCH, 0, es.evdata, cast(void*)es);
        int r = kevent(id, &ke, 1, null, 0, null);
        assert(r != -1, format("kevent(ENABLE) failed with errno %d", errno()));
    }

    void disableEvent(short evtype, EventSource es) const {
        kevent_t ke;
        // EV_SET(&ke, es.id, es.evtype, EV_DISABLE, 0, 0, cast(void*)es, [0, 0, 0, 0]);
        EV_SET(&ke, es.id, evtype, EV_DISABLE, 0, 0, cast(void*)es);
        int r = kevent(id, &ke, 1, null, 0, null);
        assert(r != -1, format("kevent(DISABLE) failed with errno %d", errno()));
    }

    private int kqueue_wait(kevent_t *earr, int len) {
        int n;
        n = kevent(id, null, 0, earr, len, null);
        if (-1 == n) {
            writefln("%s() - kevent() failed: %s", __FUNCTION__, fromStringz(strerror(errno)));
        }
        return n;
    }

    bool wait() {

        const int maxEvents = 8;
        kevent_t[maxEvents] events;

        if (done)
            return false;

        int n = kqueue_wait(events.ptr, maxEvents);
        if (-1 == n)
            return false;

        foreach (k; 0 .. n) {
            EventSource s = cast(EventSource)events[k].udata;
            ulong ecode;
            
            if (events[k].flags & (EV_ERROR | EV_EOF)) {
                ecode = 2;
            } else {
                ecode = s.eventCode(events[k]);
            }
            
            mq ~= Message(null, s.owner, ecode, s);
        }

        return true;
    }

    void stop() {
        done = true;
    }
}
